# Copyright 2016 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import patch

import ceilometer_contexts as contexts
import ceilometer_utils as utils

from test_utils import CharmTestCase, mock_open

TO_PATCH = [
    'config',
    'relation_get',
    'relation_ids',
    'related_units',
    'os_release',
]


class CeilometerContextsTest(CharmTestCase):

    def setUp(self):
        super(CeilometerContextsTest, self).setUp(contexts, TO_PATCH)
        self.config.side_effect = self.test_config.get
        self.relation_get.side_effect = self.test_relation.get

    def tearDown(self):
        super(CeilometerContextsTest, self).tearDown()

    def test_logging_context(self):
        self.test_config.set('debug', False)
        self.test_config.set('verbose', False)
        self.assertEqual(contexts.LoggingConfigContext()(),
                         {'debug': False, 'verbose': False})
        self.test_config.set('debug', True)
        self.test_config.set('verbose', False)
        self.assertEqual(contexts.LoggingConfigContext()(),
                         {'debug': True, 'verbose': False})
        self.test_config.set('debug', True)
        self.test_config.set('verbose', True)
        self.assertEqual(contexts.LoggingConfigContext()(),
                         {'debug': True, 'verbose': True})

    @patch.object(utils, 'get_shared_secret')
    def test_ceilometer_context(self, secret):
        secret.return_value = 'mysecret'
        self.assertEqual(contexts.CeilometerContext()(), {
            'port': 8777,
            'metering_secret': 'mysecret',
            'api_workers': 1,
            'metering_time_to_live': -1,
            'event_time_to_live': -1,
        })

    @patch.object(utils, 'get_shared_secret')
    def test_ceilometer_context_ttl_values(self, secret):
        secret.return_value = 'mysecret'
        self.test_config.set('metering-time-to-live', 7.776e+06)
        self.test_config.set('event-time-to-live', 7.776e+06)
        context = contexts.CeilometerContext()()
        self.assertEqual(context, {
            'port': 8777,
            'metering_secret': 'mysecret',
            'api_workers': 1,
            'metering_time_to_live': 7776000,
            'event_time_to_live': 7776000,
        })
        self.assertTrue(type(context['metering_time_to_live']) is int)
        self.assertTrue(type(context['event_time_to_live']) is int)

    def test_ceilometer_service_context(self):
        self.relation_ids.return_value = ['ceilometer-service:0']
        self.related_units.return_value = ['ceilometer/0']
        data = {
            'metering_secret': 'mysecret',
            'keystone_host': 'test'
        }
        self.test_relation.set(data)
        self.assertEqual(contexts.CeilometerServiceContext()(), data)

    def test_ceilometer_service_context_not_related(self):
        self.relation_ids.return_value = []
        self.assertEqual(contexts.CeilometerServiceContext()(), {})

    @patch('os.path.exists')
    def test_get_shared_secret_existing(self, exists):
        exists.return_value = True
        with mock_open(utils.SHARED_SECRET, u'mysecret'):
            self.assertEqual(utils.get_shared_secret(),
                             'mysecret')

    @patch('uuid.uuid4')
    @patch('os.path.exists')
    def test_get_shared_secret_new(self, exists, uuid4):
        exists.return_value = False
        uuid4.return_value = 'newsecret'
        with patch('__builtin__.open'):
            self.assertEqual(utils.get_shared_secret(),
                             'newsecret')

    @patch.object(contexts, 'determine_apache_port')
    @patch.object(contexts, 'determine_api_port')
    def test_ha_proxy_context(self, determine_api_port, determine_apache_port):
        determine_api_port.return_value = contexts.CEILOMETER_PORT - 10
        determine_apache_port.return_value = contexts.CEILOMETER_PORT - 20

        haproxy_port = contexts.CEILOMETER_PORT
        api_port = haproxy_port - 10
        apache_port = api_port - 10

        expected = {
            'service_ports': {'ceilometer_api': [haproxy_port, apache_port]},
            'port': api_port
        }
        self.assertEqual(contexts.HAProxyContext()(), expected)
